<?php

declare(strict_types=1);

namespace Drupal\lms\Controller;

use Drupal\Component\Render\FormattableMarkup;
use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Ajax\OpenModalDialogCommand;
use Drupal\Core\Cache\CacheableMetadata;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Url;
use Drupal\lms\Entity\Answer;
use Drupal\lms\Entity\Bundle\Course;
use Drupal\lms\Entity\Lesson;
use Drupal\lms\Entity\LessonStatus;
use Drupal\lms\TrainingManager;
use Drupal\user\UserInterface;

/**
 * The training controller.
 */
final class CourseController extends ControllerBase {

  use CourseControllerTrait;

  public function __construct(protected readonly TrainingManager $trainingManager) {}

  /**
   * Start the learning path.
   *
   * This page will redirect the user to the first learning path content.
   */
  public function start(Course $group) {
    try {
      $lesson_status = $this->trainingManager->getRequestedLessonStatus($group, $this->currentUser());
    }
    catch (\Exception $e) {
      $url = $this->handleError($group, $e);
      return $this->redirect($url->getRouteName(), $url->getRouteParameters());
    }
    $route_parameters = [
      'group' => $group->id(),
      'lesson_delta' => $lesson_status->getCurrentLessonDelta($group),
      'activity_delta' => $lesson_status->getCurrentActivityDelta(),
    ];

    return $this->redirect('lms.group.answer_form', $route_parameters);
  }

  /**
   * Returns lesson question answer form title.
   */
  public function activityFormTitle(Course $group, int $lesson_delta, int $activity_delta) {
    try {
      $course_status = $this->trainingManager->getCurrentCourseStatus($group, $this->currentUser());
    }
    catch (\Exception $e) {
      return $this->t('Invalid request.');
    }

    $lesson = $group->get(Course::LESSONS)->get($lesson_delta);
    if ($lesson === NULL) {
      return $this->t('Invalid lesson delta.');
    }

    $lesson_status = $this->trainingManager->loadLessonStatus($course_status->id(), $lesson->target_id);
    $activity_item = $lesson_status->get(Lesson::ACTIVITIES)->get($activity_delta);
    if ($activity_item === NULL) {
      return $this->t('Invalid activity delta.');
    }

    return $lesson->entity->label() . ' - ' . $activity_item->entity->label();
  }

  /**
   * Specific activity callback.
   */
  public function activity(
    Course $group,
    int $lesson_delta,
    int $activity_delta
  ) {
    try {
      $lesson_status = $this->trainingManager->getRequestedLessonStatus($group, $this->currentUser(), [
        'lesson' => $lesson_delta,
        'activity' => $activity_delta,
      ]);
    }
    catch (\Exception $e) {
      $url = $this->handleError($group, $e);
      return $this->redirect($url->getRouteName(), $url->getRouteParameters());
    }

    return $this->getAnswerForm($lesson_status);
  }

  /**
   * Answer form getter method.
   *
   * @return mixed
   *   Renderable array or redirect response.
   */
  private function getAnswerForm(LessonStatus $lesson_status) {
    $lesson = $lesson_status->get('lesson')->entity;

    // @todo Implement a solution to be able to set
    // lesson / training availability,
    // preferably on the Course level - lessons reference field.
    $activity_item = $lesson_status->get(Lesson::ACTIVITIES)->get($lesson_status->get('current_activity')->value);
    $activity = $activity_item->entity;
    $build['activity_build'] = $this->entityTypeManager()->getViewBuilder('lms_activity')->view($activity, 'activity');

    // Answer form.
    $answer = $this->trainingManager->loadAnswer($lesson_status, $activity);
    if ($answer === NULL) {
      $answer = $this->trainingManager->createAnswer($lesson_status, $activity);
    }
    $build['form'] = $this->entityFormBuilder()->getForm($answer, 'default', [
      'group' => $lesson_status->get('course_status')->entity->getCourse(),
      'lms_lesson' => $lesson,
      'lms_activity' => $activity,
    ]);

    return $build;
  }

  /**
   * Results page title callback.
   */
  public function resultsTitle(Course $group, UserInterface $user) {
    return $this->t('@course results for @user', [
      '@course' => $group->label(),
      '@user' => $user->label(),
    ]);
  }

  /**
   * Results page callback.
   */
  public function results(Course $group, UserInterface $user) {
    $course_status = $this->trainingManager->loadCourseStatus($group, $user, [
      'current' => TRUE,
    ]);
    if ($course_status === NULL) {
      $completion_status = $this->t('Not started');
    }
    else {
      $completion_status = $course_status->getStatusAndScore();
    }
    $build = [];

    $build['course_status'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Completion status:'),
      'status' => [
        '#type' => 'html_tag',
        '#tag' => 'span',
        '#attributes' => ['data-lms-selector' => 'course-status'],
        '#value' => $completion_status,
      ],
    ];

    if ($course_status === NULL) {
      return $build;
    }

    $cacheability = new CacheableMetadata();
    $cacheability->addCacheableDependency($course_status);

    // If the current user is a member of the course, evaluation of
    // unevaluated activities should be possible from here.
    $is_teacher = $group->getMember($this->currentUser()) !== FALSE;
    if ($is_teacher === FALSE) {
      $is_teacher = $this->currentUser()->hasPermission('administer lms');
    }
    if ($is_teacher) {
      $build['#attached']['library'] = [
        'core/drupal.dialog.ajax',
      ];
    }

    $build['lessons'] = [];
    $lesson_statuses = $this->entityTypeManager()->getStorage('lms_lesson_status')->loadByProperties([
      'course_status' => $course_status->id(),
    ]);
    foreach ($lesson_statuses as $lesson_status) {
      $lesson_status_id = $lesson_status->id();
      $build['lessons'][$lesson_status_id] = [
        '#type' => 'details',
        '#title' => new FormattableMarkup('<span data-lms-selector="lesson-status-@lsid">@status</span>', [
          '@lsid' => $lesson_status_id,
          '@status' => $this->getLessonStatusSummary($lesson_status),
        ]),
        '#open' => FALSE,
        '#attributes' => ['class' => ['lesson-score-details']],
      ];

      $build['lessons'][$lesson_status_id]['answers'] = [
        '#theme' => 'table',
        '#header' => [
          'activity' => $this->t('Activity'),
          'answered' => $this->t('Answered'),
          'evaluated' => $this->t('Evaluated'),
          'score' => $this->t('Score'),
        ],
        '#rows' => [],
        '#empty' => $this->t('This lesson has no activities'),
      ];
      $answers = $this->trainingManager->getAnswersByActivityId($lesson_status);
      foreach ($lesson_status->getActivities() as $activity) {
        $activity_id = $activity->id();

        $answered = $this->t('No');
        $evaluated = $this->t('N/A');
        $score = $this->t('N/A');
        $selector = '';
        if (\array_key_exists($activity_id, $answers)) {
          $answered = $this->t('Yes');
          if ($answers[$activity_id]->isEvaluated()) {
            $evaluated = $this->t('Yes');
            // Show score of evaluated answers only.
            $score = $answers[$activity_id]->getScore();
          }
          elseif ($is_teacher) {
            $answer_id = $answers[$activity_id]->id();
            $selector = '-' . $answer_id;
            $evaluated = [
              '#type' => 'link',
              '#title' => $this->t('No'),
              '#url' => Url::fromRoute('lms.answer.evaluate', [
                'lms_answer' => $answer_id,
                'js' => 'nojs',
              ]),
              '#attributes' => ['class' => ['use-ajax']],
            ];
          }
          else {
            $evaluated = $this->t('No');
          }
        }

        $row = [
          'activity' => $activity->label(),
          'answered' => $answered,
          'evaluated' => [
            'data' => $evaluated,
            'data-lms-selector' => 'evaluated' . $selector,
          ],
          'score' => [
            'data' => $score,
            'data-lms-selector' => 'score' . $selector,
          ],
        ];

        // Add attributes for AJAX calls.
        $build['lessons'][$lesson_status_id]['answers']['#rows'][] = $row;
      }

      $cacheability->addCacheableDependency($lesson_status);
    }

    $cacheability->applyTo($build);
    return $build;
  }

  /**
   * Open evaluation modal form.
   */
  public function openEvaluationForm(Answer $lms_answer, string $js): array|AjaxResponse {
    $use_ajax = ($js === 'ajax');

    $build = [];

    // Render activity.
    $activity = $lms_answer->getActivity();
    $build['activity_build'] = $this->entityTypeManager()->getViewBuilder('lms_activity')->view($activity, 'activity');

    // Answer evaluation form.
    $build['form'] = $this->entityFormBuilder()->getForm($lms_answer, 'evaluate', ['use_ajax' => $use_ajax]);

    if ($use_ajax) {
      $response = new AjaxResponse();
      $response->addCommand(new OpenModalDialogCommand($this->t('Evaluate @activity', [
        '@activity' => $lms_answer->getActivity()->label(),
      ]), $build, ['width' => '80%']));
      return $response;
    }

    return $build;
  }

}
