<?php

declare(strict_types=1);

namespace Drupal\lms\Form;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Entity\ContentEntityForm;
use Drupal\Core\Entity\EntityRepositoryInterface;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\lms\ActivityAnswerManager;
use Drupal\lms\Controller\CourseControllerTrait;
use Drupal\lms\Entity\CourseStatus;
use Drupal\lms\TrainingManager;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Form controller for Answer edit forms.
 */
final class AnswerForm extends ContentEntityForm {

  use CourseControllerTrait;

  /**
   * Constructs an AnswerForm object.
   */
  public function __construct(
    EntityRepositoryInterface $entity_repository,
    EntityTypeBundleInfoInterface $entity_type_bundle_info,
    TimeInterface $time,
    protected readonly ActivityAnswerManager $answerManager,
    protected readonly TrainingManager $trainingManager
  ) {
    parent::__construct($entity_repository, $entity_type_bundle_info, $time);
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity.repository'),
      $container->get('entity_type.bundle.info'),
      $container->get('datetime.time'),
      $container->get('plugin.manager.activity_answer'),
      $container->get('lms.training_manager')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $form = parent::buildForm($form, $form_state);

    // Hide revision_log_message field.
    unset($form['revision_log_message']);

    $answer = $this->entity;
    $activity = $answer->getActivity();

    // Backwards navigation.
    $form['actions']['back'] = [
      '#type' => 'submit',
      '#value' => $this->t('Back'),
      '#submit' => [
        '::backwardsNavigation',
      ],
      '#access' => FALSE,
    ];

    // Conditionally allow backwards navigation.
    $back_parameters = $this->trainingManager->getBackNavParameters($answer);
    if (\count($back_parameters) !== 0) {
      $form['actions']['back']['#access'] = TRUE;
    }

    // Add activity specific form elements.
    $plugin_id = $this->trainingManager->getActivityPluginId($activity);
    if ($plugin_id !== NULL && $this->answerManager->hasDefinition($plugin_id)) {
      $plugin = $this->answerManager->createInstance($plugin_id);
      $plugin->answeringForm($form, $this->entity);
    }

    // Actions tweaks.
    $form['actions']['submit']['#submit'] = ['::submitForm'];
    $form['actions']['submit']['#value'] = $this->t('Submit');
    $lesson_status = $this->entity->getLessonStatus();
    if ($lesson_status->getNextActivityDelta() === NULL && $lesson_status->getNextLessonDelta() === NULL) {
      $form['actions']['submit']['#value'] = $this->t('Finish course');
    }

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $activity = $this->entity->get('activity')->entity;
    $lesson_status = $this->entity->getLessonStatus();
    $lesson = $lesson_status->getLesson();
    $course_status = $lesson_status->getCourseStatus();

    $max_score = $this->trainingManager->getActivityMaxScore($lesson, $activity);

    // Determine score and evaluated property for the answer.
    $form_state->addCleanValueKey('langcode');
    $form_state->addCleanValueKey('back');
    $form_state->cleanValues();
    $this->entity->setData($form_state->getValues());

    $plugin_id = $this->trainingManager->getActivityPluginId($activity);
    if ($plugin_id !== NULL && $this->answerManager->hasDefinition($plugin_id)) {
      $plugin = $this->answerManager->createInstance($plugin_id);
      if ($plugin->evaluatedOnSave($this->entity)) {
        $score = $plugin->getScore($this->entity) * $max_score;
        $evaluated = TRUE;
      }
      else {
        // Always set unevaluated score to max to avoid blocking user from
        // going to next lessons.
        $score = $max_score;
        $evaluated = FALSE;
        $lesson_status->setEvaluated(FALSE);
      }
    }
    // No plugin, just config - static display-only activity.
    else {
      $evaluated = TRUE;
      $score = $max_score;
    }

    $this->entity
      ->setEvaluated($evaluated)
      ->setScore($score)
      ->save();

    // Set latest user activity time on Course status.
    $this->trainingManager->setLastActivityTime($course_status);

    // Proceed to the next activity.
    $next_activity_delta = $lesson_status->getNextActivityDelta();
    $next_lesson_status = NULL;
    if ($next_activity_delta === NULL) {
      $this->trainingManager->updateLessonStatus($lesson_status);
      try {
        $next_lesson_status = $this->trainingManager->getNextLessonStatus($course_status);
      }
      catch (\Exception $e) {
        $this->trainingManager->updateCourseStatus($course_status);
        $url = $this->handleError($course_status->getCourse(), $e);
        $form_state->setRedirectUrl($url);
        return;
      }

      if ($next_lesson_status !== NULL) {
        $next_lesson_status->save();
        $next_activity_delta = $next_lesson_status->get('current_activity')->value;
        $course_status->set('current_lesson_status', $next_lesson_status);
        $lesson_status = $next_lesson_status;
      }

      // We have all required properties set on CourseStatus now,
      // we can recalculate and save.
      $last_activity = (
        $next_activity_delta === NULL &&
        $next_lesson_status === NULL
      );
      $this->trainingManager->updateCourseStatus($course_status, $last_activity);
    }
    else {
      $lesson_status->set('current_activity', $next_activity_delta);
      $lesson_status->save();
      // We also need to save Course status to update last activity time.
      $course_status->save();
    }

    if ($next_activity_delta === NULL) {
      // @todo Redirect to results?
      if ($course_status->getStatus() === CourseStatus::STATUS_PROGRESS) {
        $this->messenger()->addWarning($this->t('Last activity completed, training still has unfinished activities.'));
      }
      elseif ($course_status->getStatus() === CourseStatus::STATUS_NEEDS_EVALUATION) {
        $this->messenger()->addStatus($this->t('Course completed, please wait for evaluation.'));
      }
      elseif ($course_status->getStatus() === CourseStatus::STATUS_PASSED) {
        $this->messenger()->addStatus($this->t('Course completed.'));
      }
      elseif ($course_status->addError() === CourseStatus::STATUS_FAILED) {
        $this->messenger()->addStatus($this->t('Course failed.'));
      }
      $form_state->setRedirect('entity.group.canonical', [
        'group' => $course_status->get('gid')->target_id,
      ]);
      return;
    }

    $form_state->setRedirect('lms.group.answer_form', [
      'group' => $course_status->get('gid')->target_id,
      'lesson_delta' => $lesson_status->getCurrentLessonDelta($course_status->getCourse()),
      'activity_delta' => $lesson_status->getCurrentActivityDelta(),
    ]);
  }

  /**
   * {@inheritdoc}
   */
  public function backwardsNavigation(array $form, FormStateInterface $form_state) {
    $course_status = $this->entity->getLessonStatus()->get('course_status')->entity;
    $back_parameters = $this->trainingManager->getBackNavParameters($this->entity);

    try {
      $lesson_status = $this->trainingManager->goBackwards($course_status, $back_parameters);
    }
    catch (\Exception $e) {
      $url = $this->handleError($course_status->getCourse(), $e);
      $form_state->setRedirectUrl($url);
      return;
    }

    $form_state->setRedirect('lms.group.answer_form', [
      'group' => $course_status->get('gid')->target_id,
      'lesson_delta' => $lesson_status->getCurrentLessonDelta($course_status->getCourse()),
      'activity_delta' => $lesson_status->getCurrentActivityDelta(),
    ]);
  }

}
