<?php

declare(strict_types=1);

namespace Drupal\lms\Hooks;

use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Messenger\MessengerTrait;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\group\Entity\GroupRelationshipInterface;
use Drupal\lms\Entity\Bundle\Course;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Membership form alter hooks logic.
 */
abstract class CourseMembershipHooksBase implements ContainerInjectionInterface {

  use StringTranslationTrait;
  use MessengerTrait;

  /**
   * The constructor.
   */
  public function __construct(
    protected readonly EntityTypeManagerInterface $entityTypeManager
  ) {}

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager')
    );
  }

  /**
   * Alter group relationship entity forms.
   */
  abstract public function groupRelationshipFormAlter(&$form, FormStateInterface $form_state): void;

  /**
   * Membership form validate handler.
   */
  public function membershipFormValidate(array $form, FormStateInterface $form_state) {
    $group_relationship = $form_state->getFormObject()->getEntity();

    // Roles selected - don't do anything.
    $roles = $form_state->getValue('group_roles');
    if (\is_array($roles) && \count(\array_filter($roles)) > 0) {
      return;
    }

    $class_id = (string) $form_state->getValue('class');
    if ($class_id === NULL || $class_id === '0') {
      return;
    }

    $user_id = $form_state->getValue(['entity_id', '0', 'target_id']);
    $potential_member = $this->entityTypeManager->getStorage('user')->load($user_id);
    $course = $group_relationship->getGroup();
    if ($course->isClassMember($potential_member)) {
      $form_state->setError($form['entity_id']['widget'], $this->t('Selected user is already a member of a class on this Course.'));
      return;
    }

    $this->convertMembership($group_relationship, $class_id);
  }

  /**
   * Code saver.
   */
  protected function addClassSelector(array &$form, Course $course, bool $self_join = FALSE): void {
    $classes = $course->getClasses();
    foreach ($classes as $class) {
      $options[$class->id()] = $class->label();
    }
    if (!$self_join) {
      $options['0'] = $this->t('None (Course membership)');
    }
    if (\count($options) === 1) {
      $form['class'] = [
        '#type' => 'value',
        '#value' => \array_keys($options)[0],
      ];
      return;
    }

    $form['class'] = [
      '#type' => 'select',
      '#title' => $this->t('Select target class'),
      '#description' => $self_join ? $this->t('You will be added to the selected class.') : $this->t('Student will be added to the selected class.'),
      '#options' => $options,
    ];
  }

  /**
   * Group membership or request form submit handler.
   */
  public static function redirectToCourse(array $form, FormStateInterface $form_state): void {
    $relationship = $form_state->getFormObject()->getEntity();

    // Redirect only if self - adding / requesting.
    if ($relationship->getOwnerId() === $relationship->getEntityId()) {
      $group_id = $form_state->get('redirect_group_id');
      $form_state->setRedirect('entity.group.canonical', ['group' => $group_id]);
    }
  }

  /**
   * Display a status message.
   */
  public function addedToClassMessage(array $form, FormStateInterface $form_state): void {
    $membership = $form_state->getFormObject()->getEntity();
    $group = $membership->getGroup();
    if ($group->bundle() === 'lms_class') {
      $this->messenger()->addStatus($this->t('Student @student has been added to @class.', [
        '@student' => $membership->getEntity()->label(),
        '@class' => $group->label(),
      ]));
    }
  }

  /**
   * Converts membership from Course to Class.
   */
  private function convertMembership(GroupRelationshipInterface $group_relationship, string $class_id): void {
    $group_relationship->set('gid', $class_id);
    $types = $this->entityTypeManager->getStorage('group_relationship_type')->loadByProperties([
      'group_type' => 'lms_class',
      'content_plugin' => 'group_membership',
    ]);
    $type = \reset($types);
    $group_relationship->set('type', $type->id());
    $group_relationship->set('group_type', 'lms_class');
  }

}
