<?php

declare(strict_types=1);

namespace Drupal\lms\Hooks;

use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\group\Entity\GroupInterface;
use Drupal\lms\Entity\Bundle\Course;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Centralized entity hooks logic.
 */
final class EntityHooks implements ContainerInjectionInterface {

  /**
   * The constructor.
   */
  public function __construct(
    protected readonly EntityTypeManagerInterface $entityTypeManager
  ) {}

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager')
    );
  }

  /**
   * Act on inserting a group.
   */
  public function groupInsert(GroupInterface $group) {
    if (!$group instanceof Course) {
      return;
    }

    // Add a default class on learning path creation.
    $class = $this->entityTypeManager->getStorage('group')->create([
      'type' => 'lms_class',
      'label' => self::generateDefaultClassName($group),
      'uid' => $group->getOwnerId(),
    ]);
    $class->save();

    // Always add the creator membership.
    $group_type = $class->getGroupType();
    if ($group_type->creatorMustCompleteMembership()) {
      $values = ['group_roles' => $group_type->getCreatorRoleIds()];
      $class->addMember($class->getOwner(), $values);
    }

    $group->addRelationship($class, 'lms_classes');
  }

  /**
   * Generate default class name.
   */
  public static function generateDefaultClassName(Course $course): string {
    // Generate a class name based on the learning path. Using the learning
    // path name could make group listings hard to scan. Using the learning
    // path ID could create confusion with the class's own numeric ID. To
    // avoid this and also avoid hash collisions, add 100 to the learning
    // path ID then base36 encode and upper-case it. This gives us a
    // human-readable short hash which is superficially similar to the
    // kind of class names that are often used in schools (1A, DF1, etc.).
    return (string) \t('Class @class_name', [
      '@class_name' => \strtoupper(\base_convert((string) ($course->id() + 100), 10, 36)),
    ]);
  }

}
