<?php

declare(strict_types=1);

namespace Drupal\lms\Plugin\Field\FieldType;

use Drupal\Core\Field\Attribute\FieldType;
use Drupal\Core\Field\FieldStorageDefinitionInterface;
use Drupal\Core\Field\Plugin\Field\FieldType\EntityReferenceItem;
use Drupal\Core\Field\Plugin\Field\FieldType\MapItem;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\TypedData\MapDataDefinition;

/**
 * Entity reference with additional data.
 */
#[FieldType(
  id: 'lms_reference',
  label: new TranslatableMarkup('LMS Entity reference'),
  description: new TranslatableMarkup('An entity field containing an entity reference with additional serialized data.'),
  category: 'lms',
  default_widget: 'lms_reference',
  default_formatter: 'lms_reference_label',
  list_class: '\Drupal\Core\Field\EntityReferenceFieldItemList',
)]
final class LMSReferenceItem extends EntityReferenceItem {

  /**
   * Check if this step is mandatory.
   *
   * Currently only used if lesson is referenced.
   */
  public function isMandatory(): bool {
    if (!\array_key_exists('mandatory', $this->data)) {
      return FALSE;
    }
    return (bool) $this->data['mandatory'];
  }

  /**
   * Get minimum success score for the step.
   *
   * Currently only used if lesson is referenced.
   */
  public function getRequiredScore(): int {
    if (!\array_key_exists('required_score', $this->data)) {
      return 0;
    }
    return (int) $this->data['required_score'];
  }

  /**
   * Get maximum score.
   *
   * Currently only used if activity is referenced.
   */
  public function getMaxScore(): int {
    if (!\array_key_exists('max_score', $this->data)) {
      return 0;
    }
    return (int) $this->data['max_score'];
  }

  /**
   * {@inheritdoc}
   */
  public static function propertyDefinitions(FieldStorageDefinitionInterface $field_definition) {
    $definitions = parent::propertyDefinitions($field_definition);
    $definitions['data'] = MapDataDefinition::create();
    return $definitions;
  }

  /**
   * {@inheritdoc}
   */
  public static function schema(FieldStorageDefinitionInterface $field_definition) {
    $schema = parent::schema($field_definition);
    $schema['columns']['data'] = [
      'type' => 'blob',
      'size' => 'big',
      'serialize' => TRUE,
    ];
    return $schema;
  }

  /**
   * {@inheritdoc}
   */
  public function setValue($values, $notify = TRUE) {
    if (\array_key_exists('data', $values)) {
      if (!\is_array($values['data'])) {
        if ($values instanceof MapItem) {
          $values['data'] = $values->getValue();
        }
        elseif (\is_string($values['data'])) {
          $values['data'] = \unserialize($values['data'], ['allowed_classes' => FALSE]);
        }
      }
      $this->values['data'] = $values['data'];
    }

    parent::setValue($values, $notify);
  }

}
