<?php

namespace Drupal\lms\Plugin\Field\FieldWidget;

use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Ajax\InvokeCommand;
use Drupal\Core\Entity\Element\EntityAutocomplete;
use Drupal\Core\Field\Attribute\FieldWidget;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Field\Plugin\Field\FieldWidget\EntityReferenceAutocompleteWidget;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;

/**
 * Plugin implementation of the 'entity_reference_autocomplete' widget.
 */
#[FieldWidget(
  id: 'lms_reference',
  label: new TranslatableMarkup('LMS reference'),
  description: new TranslatableMarkup('Entity autocomplete with additional data.'),
  field_types: ['lms_reference'],
)]
class LMSReferenceItemWidget extends EntityReferenceAutocompleteWidget {

  /**
   * {@inheritdoc}
   */
  public function formElement(FieldItemListInterface $items, $delta, array $element, array &$form, FormStateInterface $form_state) {
    $element = parent::formElement($items, $delta, $element, $form, $form_state);

    $element['data'] = [];

    $data = $items[$delta]->data;
    if ($data === NULL) {
      $data = [];
    }

    $element['target_id']['#title_display'] = 'before';

    $target_type = $this->fieldDefinition->getSetting('target_type');
    if ($target_type === 'lms_activity') {
      $element['target_id']['#title'] = $this->t('Activity');
      $element['data']['max_score'] = [
        '#type' => 'number',
        '#min' => 0,
        '#max' => 100,
        '#title' => $this->t('Maximum score'),
        '#default_value' => 5,
      ];
      $element['target_id']['#ajax'] = [
        'callback' => [\get_class($this), 'updateMaxScore'],
        'event' => 'autocompleteclose',
      ];
    }
    elseif ($target_type === 'lms_lesson') {
      $element['target_id']['#title'] = $this->t('Lesson');
      $element['data']['mandatory'] = [
        '#type' => 'checkbox',
        '#title' => $this->t('Mandatory'),
        '#default_value' => TRUE,
      ];
      $element['data']['required_score'] = [
        '#type' => 'number',
        '#min' => 0,
        '#max' => 100,
        '#title' => $this->t('Required score [%]'),
        '#default_value' => 50,
      ];
    }

    // Apply default values.
    foreach (\array_keys($element['data']) as $sub_property) {
      if (\array_key_exists($sub_property, $data)) {
        $element['data'][$sub_property]['#default_value'] = $data[$sub_property];
      }
    }

    $element['data']['#tree'] = TRUE;
    $element['data']['#weight'] = 100;

    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public function massageFormValues(array $values, array $form, FormStateInterface $form_state) {
    $values = parent::massageFormValues($values, $form, $form_state);
    return $values;
  }

  /**
   * Update max score value on new elements.
   */
  public static function updateMaxScore(array $form, FormStateInterface $form_state): ?AjaxResponse {
    $trigger = $form_state->getTriggeringElement();
    $entity_id = EntityAutocomplete::extractEntityIdFromAutocompleteInput($trigger['#value']);
    if ($entity_id === NULL) {
      return NULL;
    }
    $entity_type_manager = \Drupal::entityTypeManager();
    $entity = $entity_type_manager->getStorage('lms_activity')->load($entity_id);
    if ($entity === NULL) {
      return NULL;
    }

    // Value.
    $activity_type = $entity_type_manager->getStorage('lms_activity_type')->load($entity->bundle());
    $default_max_score = $activity_type->getDefaultMaxScore();

    // Selector.
    $parents = $trigger['#parents'];
    \array_pop($parents);
    $parents[] = 'data';
    $parents[] = 'max_score';
    $selector = 'input[name="' . \array_shift($parents) . '[' . \implode('][', $parents) . ']"]';

    // Command.
    $response = new AjaxResponse();
    $response->addCommand(new InvokeCommand($selector, 'val', [$default_max_score]));
    return $response;
  }

}
